/*
  CPU IO board connections:
   1 A7 NC           21 ~Reset Switch 
   2 A6 NC           33 GND
   3 A5 NC           34 D7
   4 A4 NC           35 D6
   5 A3 NC           36 D5
   6 A2 DECODED      37 D4
   7 A1 DECODED      38 D3
   8 A0 DECODED      39 D2
   9 VCC             40 D1
  10 ~RD             41 D0
  11 ~WR           
  12 ~REGS         

  Addressing:
    Regs = Effectively 0XF8 to 0XFF
    Addr = Effectively 0XFA
    Data = Effectively 0XFB

  Notes: Use on board power as Arduino is not sufficient

  CPU Emulation:
    This CPU has no actual opcodes.
    All instructions are MOVE from [PC] to [PC+1]
    i.e. Move data from source address to destination address.

    The 4 cycles are:
    1 Fetch the destination address pointed to by the program counter and put in the Addr register
    2 Fetch desination data pointed to by the Addr register and put and the Data register
    2A  Increment the program counter (PC)
    3 Fetch destination address pointed to by the program counter and put in the Addr register
    4 Deposit the Data register into the destination pointed to by the Addr register
    4A  Increment the program counter (PC)
    
    All hardware (PC,IO and ALU) is memory mapped.
    
    Arduino pinouts:
      D0 -> D0
      D1 -> D1
      D2 -> D2
      D3 -> D3
      D4 -> D4
      D5 -> D5
      D6 -> D6
      D7 -> D7
      D8 -> A0
      D9 -> A1
      D10 -> A2
      D11 -> REG
      D12 -> WR
      D13 -> RD
*/
#include <Arduino.h>
#include <avr/pgmspace.h>

void Put_Data(byte Data){
  // Write the Data Register (Yellow)
  DDRD  = B11111111;    // Set databus to output
  PORTD = Data;         // Write the data
  PORTB = B00100011;    // WR low
  delayMicroseconds(1); // Let the bus settle
  PORTB = B00110011;    // WR high
  DDRD  = B00000000;    // Set databus to input
  PORTD = B00000000;    // Turn off pull up resistors (i.e. not used)
}

void Put_Addr(byte Data){
  // Write the Addr Register (Red)
  DDRD  = B11111111;    // Set databus to output
  PORTD = Data;         // Write the data
  PORTB = B00100010;    // WR low
  delayMicroseconds(1); // Let the bus settle
  PORTB = B00110010;    // WR high
  DDRD  = B00000000;    // Set databus to input
  PORTD = B00000000;    // Turn off pull up resistors (i.e. not used)
}

byte Get_Data(void){
  byte Data;
  // Read the Data Register
  DDRD  = B00000000;    // Set databus to input
  PORTD = B00000000;    // Turn off pull up resistors (i.e. not used)
  PORTB = B00010011;    // RD low 
  delayMicroseconds(1); // Let the bus settle
  Data = PIND;          // Read Port D
  PORTB = B00110011;    // RD high
  return Data;
}
byte Get_Addr(void){
  byte Addr;
  // Read the Addr Register
  DDRD  = B00000000;    // Set databus to input
  PORTD = B00000000;    // Turn off pull up resistors (i.e. not used)
  PORTB = B00010010;    // RD low 
  delayMicroseconds(1); // Let the bus settle
  Addr = PIND;          // Read Port D
  PORTB = B00110010;    // RD high
  return Addr;
}

void test_io(void){
  while(true) {
    byte Data;
    byte Addr;
    
    // Read and write Addr and Data switches
    Data=Get_Data();
    Put_Data(~Data);
    Addr=Get_Addr();
    Put_Addr(~Addr);
    delay(100);
  }
}

void SingleStep(byte Data, byte Addr){
  Put_Data(Data);
  Put_Addr(Addr);
  while (digitalRead(A0)==HIGH);
  delay(10);
  while (digitalRead(A0)==LOW);
  delay(10);
}  

void setup(){
  // Initialise control pins on Port B
  // Bit   76543210
  DDRB  = B00111111;
  PORTB = B00110010;
  // PortB xxRWRAAA
  // PortB xxDRx210
  
  // Read the Reset pin
  pinMode(A0,INPUT);
}

const byte PROGMEM ROM[128] ={
// Monitor
  0X7F, 0XF2,  //  S0: 0X00  _128 -> ADDR
  0XF2, 0XFA,  //  L0: 0X02  ADDR -> A
  0X7E, 0XF7,  //      0X04  _255 -> JUMP
  0X7D, 0XF6,  //      0X06  _241 -> RPTR
  0X7C, 0XF5,  //      0X08  _243 -> DEPOSIT
  0XF2, 0XF4,  //      0X0A  ADDR -> FETCH
  0X77, 0XF1,  //      0X0C  L1 -> RETURN
  0X7B, 0XFF,  //      0X0E  _244 -> JMP
  0XF3, 0XFB,  //  L1: 0X10  DATA -> D
  0XFA, 0XF8,  //      0X12  A -> P
  0XFA, 0XF9,  //      0X14  A -> Q
  0XF9, 0XF8,  //      0X16  NAND -> P
  0XF2, 0XF9,  //      0X18  ADDR -> Q
  0XF9, 0XF9,  //      0X1A  NAND -> Q
  0XF9, 0XF0,  //      0X1C  NAND -> TMP
  0XF2, 0XF8,  //      0X1E  ADDR -> P
  0XF9, 0XF9,  //      0X20  NAND -> Q
  0XF0, 0XF8,  //      0X22  TMP -> P
  0XF9, 0XF2,  //      0X24  NAND -> ADDR
  0XFB, 0XF8,  //      0X26  D -> P
  0XFB, 0XF9,  //      0X28  D -> Q
  0XF9, 0XF8,  //      0X2A  NAND -> P
  0XF3, 0XF9,  //      0X2C  DATA -> Q
  0XF9, 0XF9,  //      0X2E  NAND -> Q
  0XF9, 0XF0,  //      0X30  NAND -> TMP
  0XF3, 0XF8,  //      0X32  DATA -> P
  0XF9, 0XF9,  //      0X34  NAND -> Q
  0XF0, 0XF8,  //      0X36  TMP -> P
  0XF9, 0XF3,  //      0X38  NAND -> DATA
  0XFB, 0XF8,  //      0X3A  D -> P
  0X7A, 0XF9,  //      0X3C  _1 -> Q
  0X76, 0XFE,  //      0X3E  L2 -> JC
  0XF2, 0XF8,  //      0X40  ADDR -> P
  0X7A, 0XF9,  //      0X42  _1 -> Q
  0X7F, 0XFE,  //      0X44  _128 -> JC
  0X7E, 0XF7,  //      0X46  _255 -> JUMP
  0X7D, 0XF6,  //      0X48  _241 -> RPTR
  0XF2, 0XF5,  //      0X4A  ADDR -> DEPOSIT
  0X7C, 0XF4,  //      0X4C  _243 -> FETCH
  0X76, 0XF1,  //      0X4E  L2 -> RETURN
  0X7B, 0XFF,  //      0X50  _244 -> JMP
  0XFA, 0XF8,  //  L2: 0X52  A -> P
  0XFB, 0XF9,  //      0X54  D -> Q
  0XF9, 0XF8,  //      0X56  NAND -> P
  0X7E, 0XF9,  //      0X58  _255 -> Q
  0X76, 0XFE,  //      0X5A  L2 -> JC
  0X75, 0XF1,  //  L3: 0X5C  L0 -> RETURN
  0X79, 0XF0,  //      0X5E  _10 -> DELAY
  0XF0, 0XF8,  //  D0: 0X60  DELAY -> P
  0X7E, 0XF9,  //      0X62  _255 -> Q
  0X74, 0XFE,  //      0X64  D1 -> JC
  0XF1, 0XFF,  //      0X66  RETURN -> JMP
  0XF8, 0XF0,  //  D1: 0X68  ADD -> DELAY
  0X78, 0XF8,  //      0X6A  _123 -> P
  0XF8, 0XF8,  //  D2: 0X6C  ADD -> P
  0X73, 0XFE,  //      0X6E  D2 -> JC
  0X72, 0XFF,  //      0X70  D0 -> JMP
  0X60, 0X6C,  //      0X72  Const,   Const
  0X68, 0X02,  //      0X74  Const,   Const
  0X52, 0X10,  //      0X76  Const,   Const
  0X7B, 0X0A,  //      0X78  Const,   Const
  0X01, 0XF4,  //      0X7A  Const,   Const
  0XF3, 0XF1,  //      0X7C  Const,   Const
  0XFF, 0X80   //      0X7E  Const,   Const
};
byte RAM[120]={
// LED light chaser
  0XEF, 0XF2,  //  P0: 0X80  _1 -> V2
  0XEE, 0XF3,  //      0X82  _128 -> V3
  0XF2, 0XFB,  //  L0: 0X84  V2 -> D
  0XF3, 0XFA,  //      0X86  V3 -> A
  0XF2, 0XF8,  //      0X88  V2 -> P
  0XF2, 0XF9,  //      0X8A  V2 -> Q
  0XEA, 0XFE,  //      0X8C  L1 -> JC
  0XF8, 0XF2,  //      0X8E  ADD -> V2
  0XE9, 0XFF,  //      0X90  L2 -> JMP
  0XEF, 0XF2,  //  L1: 0X92  _1 -> V2
  0XED, 0XF4,  //  L2: 0X94  _7 -> V4
  0XF3, 0XF8,  //  L3: 0X96  V3 -> P
  0XF3, 0XF9,  //      0X98  V3 -> Q
  0XE8, 0XFE,  //      0X9A  L4 -> JC
  0XF8, 0XF3,  //      0X9C  ADD -> V3
  0XE7, 0XFF,  //      0X9E  L5 -> JMP
  0XEF, 0XF3,  //  L4: 0XA0  _1 -> V3
  0XF4, 0XF8,  //  L5: 0XA2  V4 -> P
  0XEC, 0XF9,  //      0XA4  _255 -> Q
  0XF8, 0XF4,  //      0XA6  ADD -> V4
  0XF8, 0XF8,  //      0XA8  ADD -> P
  0XE6, 0XFE,  //      0XAA  L3 -> JC
  0XEC, 0XF0,  //      0XAC  _255 -> DELAY
  0XE5, 0XF1,  //      0XAE  L0 -> RETURN
  0XEB, 0XFF,  //      0XB0  _96 -> JMP
  0X00, 0X00,  //      0XB2  NOP,   NOP
  0X00, 0X00,  //      0XB4  NOP,   NOP
  0X00, 0X00,  //      0XB6  NOP,   NOP
  0X00, 0X00,  //      0XB8  NOP,   NOP
  0X00, 0X00,  //      0XBA  NOP,   NOP
  0X00, 0X00,  //      0XBC  NOP,   NOP
  0X00, 0X00,  //      0XBE  NOP,   NOP
  0X00, 0X00,  //      0XC0  NOP,   NOP
  0X00, 0X00,  //      0XC2  NOP,   NOP
  0X00, 0X00,  //      0XC4  NOP,   NOP
  0X00, 0X00,  //      0XC6  NOP,   NOP
  0X00, 0X00,  //      0XC8  NOP,   NOP
  0X00, 0X00,  //      0XCA  NOP,   NOP
  0X00, 0X00,  //      0XCC  NOP,   NOP
  0X00, 0X00,  //      0XCE  NOP,   NOP
  0X00, 0X00,  //      0XD0  NOP,   NOP
  0X00, 0X00,  //      0XD2  NOP,   NOP
  0X00, 0X00,  //      0XD4  NOP,   NOP
  0X00, 0X00,  //      0XD6  NOP,   NOP
  0X00, 0X00,  //      0XD8  NOP,   NOP
  0X00, 0X00,  //      0XDA  NOP,   NOP
  0X00, 0X00,  //      0XDC  NOP,   NOP
  0X00, 0X00,  //      0XDE  NOP,   NOP
  0X00, 0X00,  //      0XE0  NOP,   NOP
  0X00, 0X00,  //      0XE2  NOP,   NOP
  0X00, 0X84,  //      0XE4  NOP,   Const
  0X96, 0XA2,  //      0XE6  Const,   Const
  0XA0, 0X94,  //      0XE8  Const,   Const
  0X92, 0X60,  //      0XEA  Const,   Const
  0XFF, 0X07,  //      0XEC  Const,   Const
  0X80, 0X01   //      0XEE  Const,   Const
};  
// P and Q are integers in order to test for carry
int  RegP=0;   // An ALU register at 0xf9
int  RegQ=0;   // An ALU register at 0xf9
byte Addr=0;   // The address register
byte Data=0;   // The data register
byte PC=0;     // The program counter
void loop(){

  // Cycle 1 - Get the fetch address
  if (PC<0x80) {
    Addr=pgm_read_byte(ROM+PC);     // The ROM is stored in program space
  } else if (PC<0xf8) {
    Addr=RAM[PC-128]; 
  } else if (PC==0xff) {            // Read the JMP Register?
    Addr=0xff;
  } else if (PC==0xfe) {            // Read the JC Register?
    Addr=0xff;
  } else if (PC==0xfd) {            // Read the JEQ Register?
    Addr=0xff;
  } else if (PC==0xfc) {            // Read the JGT Register?
    Addr=0xff;
  } else if (PC==0xfb) {            // Read the Data Register
    Addr=Get_Data();
  } else if (PC==0xfa) {            // Read the Addr Register
    Addr=Get_Addr();
  } else if (PC==0xf9) {            // Read the NAND Register
    Addr=~(RegP&RegQ);
  } else if (PC==0xf8) {            // Read the ADD Register
    Addr=RegP+RegQ;
  } 
  // if ((PC>=128)&&(PC<=239)) SingleStep(Addr,PC);  
  
  // Cycle 2 - Fetch the data
  if (Addr<0x80) {
    Data=pgm_read_byte(ROM+Addr);   // The ROM is stored in program space
  } else if (Addr<0xf8) {
    Data=RAM[Addr-128]; 
  } else if (Addr==0xff) {          // Read the JMP Register?
    Data=0xff;
  } else if (Addr==0xfe) {          // Read the JC Register?
    Data=0xff;
  } else if (Addr==0xfd) {          // Read the JEQ Register?
    Data=0xff;
  } else if (Addr==0xfc) {          // Read the JGT Register?
    Data=0xff;
  } else if (Addr==0xfb) {          // Read the Data Register
    Data=Get_Data();
  } else if (Addr==0xfa) {          // Read the Addr Register
    Data=Get_Addr();
  } else if (Addr==0xf9) {          // Read the NAND Register
    Data=~(RegP&RegQ);
  } else if (Addr==0xf8) {          // Read the ADD Register
    Data=RegP+RegQ;
  }
  // if ((PC>=128)&&(PC<=239)) SingleStep(Data,PC); 

  // Increment the program counter
  PC++;
  
  // Cycle 3 - Get the deposit address
  if (PC<0x80) {
    Addr=pgm_read_byte(ROM+PC);     // The ROM is stored in program space
  } else if (PC<0xf8) {
    Addr=RAM[PC-128]; 
  } else if (Addr<0xf8) {
    Addr=RAM[Addr]; 
  } else if (PC==0xff) {            // Read the JMP Register?
    Addr=0xff;
  } else if (PC==0xfe) {            // Read the JC Register?
    Addr=0xff;
  } else if (PC==0xfd) {            // Read the JEQ Register?
    Addr=0xff;
  } else if (PC==0xfc) {            // Read the JGT Register?
    Addr=0xff;
  } else if (PC==0xfb) {            // Read the Data Register
    Addr=Get_Data();
  } else if (PC==0xfa) {            // Read the Addr Register
    Addr=Get_Addr();
  } else if (PC==0xf9) {            // Read the NAND Register
    Addr=~(RegP&RegQ);
  } else if (PC==0xf8) {            // Read the ADD Register
    Addr=RegP+RegQ;
  } 

  // Cycle 4 - Deposit the data
  if (Addr<0x80) {
    // ROM[Addr]=Data;
  } else if (Addr<0xf8) {
    RAM[Addr-128]=Data; 
  } else if (Addr==0xff) {          // JMP Register
  } else if (Addr==0xfe) {          // JC Register
  } else if (Addr==0xfd) {          // JEQ Register
  } else if (Addr==0xfc) {          // JGT Register
  } else if (Addr==0xfb) {          // Write the Data Register
    Put_Data(Data);
  } else if (Addr==0xfa) {          // Write the Addr Register
    Put_Addr(Data);
  } else if (Addr==0xf9) {          // Write the Q Register
    RegQ=Data;
  } else if (Addr==0xf8) {          // Write the P Register
    RegP=Data;
  }
  // if ((PC>=128)&&(PC<=239)) SingleStep(Addr,PC); 
  
  // Increment the program counter
  if (Addr==0xff) {                 // Write the JMP Register
    PC=Data;
  } else if (Addr==0xfe) {          // Write the JC Register
    if (RegP+RegQ>255) {
      PC=Data;
    } else {
      PC++;
    }  
  } else if (Addr==0xfd) {          // Write the JEQ 
    if (RegP==RegQ) {
      PC=Data;
    } else {
      PC++;
    }     
  } else if (Addr==0xfc) {          // Write the JGT Register
    if (RegP>RegQ) {
      PC=Data;
    } else {
      PC++;
    } 
  } else {
    PC++;
  }


}

    
    
    
